#
# hashchain-fishing.R, 31 Mar 25
#
# Data from:
# Efficient Exact Online String Matching Through Linked Weak Factors
# Matthew N. Palmer, Simone Faro, and Stefano Scafiti

library("colorspace")
library("plyr")

library("PlackettLuce")

par(bty="l")
par(las=1)
par(pch="+")
pal_col=rainbow(2)


alg_str="HC3"

# Plot the density of benchmark times for the PLEN values
plot_alg_density=function(df)
{
plot(0, type="n",
	yaxs="i",
	xlim=c(0.3, 10), ylim=c(0, 10.1),
	xlab="Time (ms)", ylab="Density")

d_ply(df, .(PLEN), function(X) lines(density(X$stime)))

}


time_vs_plen=function(alg_name)
{
exalg=subset(hc, ALGORITHM == alg_name) # pick one

plot(exalg$PLEN, exalg$stime, log="xy", main=alg_name,
	xlab="Pattern length", ylab="Time (total, ms)")

return(NULL)
}


# Build rank of algorithms
rank_run=function(df)
{
alg=df[order(df$ALGORITHM), ] # Ensure consistent order

alg_rank=as.data.frame(t(rank(alg$stime))) # build colum of rankings
colnames(alg_rank)=alg$ALGORITHM

return(alg_rank)
}


# Fit PlackettLuce model
fit_PL=function(Nlen, run_cnt=500)
{
len_N=subset(hc, PLEN == Nlen)
alg_cnt=count(len_N$ALGORITHM)
# Only interested in algorithms run 500 times
s_500=subset(alg_cnt, freq == run_cnt)
len_N=subset(len_N, ALGORITHM %in% s_500$x)

alg_rank=ddply(len_N, .(run), rank_run)
alg_rank$run=NULL

# pl=PlackettLuce(alg_rank) # "iterative scaling" fails to converge
pl=PlackettLuce(alg_rank, method="BFGS")

return(pl)
}


alg_covar_info=function(df, list_of_algs)
{
a_index=grepl(paste0("^", df$ALGORITHM, "[0-9]*$"), list_of_algs)

m_alg=list_of_algs[a_index]

comb_info=df[rep(1, length(m_alg)), ]
comb_info$ALGORITHM=m_alg
return(comb_info)
}


# How much of the variance is explained by the fitted model?
# McFadden’s R²
McFadden=function(the_rank, the_pl)
{
# The following call will fail unless the items are 'connected'.
# That is they are all both greater-than and less-than some items.
# An algorithm that is always the fastest/slowest is not connected.
# t=connectivity(the_rank)
# which(t$membership == 1) # if always ranked first
all_null=PlackettLuce(the_rank, method="BFGS", weights=rep(1, nrow(the_rank)) , npseudo=0)
ll_full=logLik(the_pl)
ll_null=logLik(all_null)
r2_mcfadden <- 1 - (ll_full / ll_null)

return(r2_mcfadden)
}


# Initial version from Grok, and then fixed up
nagelkerke_r2 = function(pl_model)
{
# Extract log-likelihood of the fitted model
LL_fitted = as.numeric(logLik(pl_model))
  
# Extract the number of observations (n)
# n = pl_model$nobs
n=nrow(pl_model$rankings)

# Null model: assumes equal worth for all items
# Create a null model by setting all worths to 1 (log-worth = 0)
null_data = pl_model$rankings  # Use the same ranking data
null_model = PlackettLuce(rankings = null_data, npseudo = 0, method="BFGS")
  
# Extract log-likelihood of the null model
LL_null = as.numeric(logLik(null_model))
  
# Calculate Nagelkerke's R²
# Formula: R²_N = [1 - (L_null / L_fitted)^(2/n)] / [1 - L_null^(2/n)]
# L_null = exp(LL_null)    # Convert log-likelihood to likelihood
# L_fitted = exp(LL_fitted)
# nagelkerke_r2 = (1 - (L_null / L_fitted)^(2 / n)) / (1 - L_null^(2 / n))
# We don't want large values to overflow
nagelkerke_r2 = (1 - exp(LL_null - LL_fitted)^(2 / n)) / (1 - exp(LL_null*(2 / n)))
  
return(nagelkerke_r2)
}


# hc=read.csv("DNA1683734900_m.csv", sep="\t")
hc=read.csv("TXT1683705119_m.csv", sep="\t")
alg_info=read.csv("alg_info.csv")


hc$run=as.numeric(hc$MEASUREMENT)
hc$stime=as.numeric(hc$TOTAL.TIME..ms.)

exalg=subset(hc, ALGORITHM == alg_str) # pick one
ex4=subset(exalg, PLEN == 4)
ex4=subset(exalg, PLEN == 4)
ex256=subset(exalg, PLEN == 256)


plot_alg_density(exalg)


ex_mod=glm(log(stime) ~ log(PLEN), data=exalg)
summary(ex_mod)

plot(exalg$PLEN, exalg$stime, log="xy",
	xlab="Pattern length", ylab="Time (total, ms)")

dev.copy(dev=pdf, file=paste0(alg_str, ".pdf"))
dev.off()


exalg=subset(hc, ALGORITHM == "BSDM3") # pick one

plot(exalg$PLEN, exalg$stime, log="xy", col=pal_col[2],
	ylim=c(0.5, 60),
	xlab="Pattern length", ylab="Time (total, ms)")
exalg=subset(hc, ALGORITHM == "HC3") # pick one

points(exalg$PLEN*1.1, exalg$stime, col=pal_col[1])

legend("topright", legend=c("HC3", "BSDM3"), cex=1.2,
                        fill=pal_col, border="white", bty="n")

dev.copy(dev=png, file="HC-patlen-time.png")
dev.off()


par(mfcol=c(3, 4))
time_vs_plen("BNDMQ4")
time_vs_plen("BSDM3")
time_vs_plen("BXS2")
time_vs_plen("FHC2")
time_vs_plen("FJS")
time_vs_plen("HASH3")
time_vs_plen("HC3")
time_vs_plen("QF34")
time_vs_plen("THC3")
time_vs_plen("TWFR4")
time_vs_plen("UFM4")
time_vs_plen("WFR2")


dev.copy(dev=pdf, file="random_algs.pdf")
dev.off()

par(mfcol=c(1, 1))


# Test cases
len_N=subset(hc, PLEN == 8)
alg_cnt=count(len_N$ALGORITHM)
# Only interested in algorithms run 500 times
s_500=subset(alg_cnt, freq == 500)
len_N=subset(len_N, len_N$ALGORITHM %in% s_500$x)

alg_rank=ddply(len_N, .(run), rank_run)
alg_rank$run=NULL

# pl=PlackettLuce(alg_rank) # "iterative scaling" fails to converge
pl=PlackettLuce(alg_rank, method="BFGS")

summary(pl)
nagelkerke_r2(pl)

plot(qvcalc(pl))

# End of test cases


# Fit model for PLEN values 2^(3:9)
dum=sapply(2^(3:9), function(X) {pl8=fit_PL(X); print(c(X, sort(coef(pl8), decreasing=TRUE)))})

alg_cnt=count(hc$ALGORITHM)
# Only interested in algorithms run 500 times
s_500=subset(alg_cnt, freq == 11*500)
len_N=subset(hc, ALGORITHM %in% s_500$x)


# Aggregate ranking for every PLEN and fit a single model
all_rank=ddply(len_N, .(PLEN), function(df) ddply(df, .(run), rank_run))
all_rank$PLEN=NULL
all_rank$run=NULL

all_pl=PlackettLuce(all_rank, method="BFGS")
sort(coef(all_pl), decreasing=TRUE) # print the lambda values


# How much of the variance is explained by the fitted model?
# McFadden’s R² is likely to be tiiny because there are so many algorithms
McFadden(all_rank, all_pl) # -3.530103e-08 not a lot!
nagelkerke_r2(all_pl) # 3.833385e-06 still really small


# plot(qvcalc(all_pl))
# Draw our own plot until qvcalc plot is improved per submitted GitHub issue
# Get estimats and standard errors
t=qvcalc(all_pl)
bt=data.frame(ability=t$qvframe$estimate, se=t$qvframe$SE)
rownames(bt)=rownames(t$covmat)
bt=bt[order(bt$ability), ] # sort

# Base plot
plot(-5, type="n",
        xlim=c(0.0, 10.0), ylim=c(1, nrow(bt)),
        yaxt="n",
        xlab=expression(lambda), ylab="")
axis(2, at=1:nrow(bt), label=rownames(bt))

# Add line for value and standard error
dum=sapply(1:nrow(bt), function(X)
                        {
                        lines(c(-bt$se[X], bt$se[X])+bt$ability[X], c(X, X), col=pal_col[2], lwd=3.5)
                        lines(c(bt$ability[X], bt$ability[X]), c(-0.1, 0.1)+X, col=pal_col[1], lwd=2.5)
                        })

dev.copy(dev=png, file="HC-all-lambda.png")
dev.off()


# Build covariate dataframe
alg_covar=ddply(alg_info, .(ALGORITHM), alg_covar_info, s_500$x)

# Model impact of covariates
pl_P=pladmm(all_rank, ~ automata+char_compare+weak_factor+sentinel, data=alg_covar)

